@@ -496,7 +496,7 @@ struct pattern *pat_match_str(struct sample *smp, struct pattern_expr *expr, int
 	}
 
 	/* look in the list */
-	if (pat_lru_tree && !LIST_ISEMPTY(&expr->patterns)) {
+	if (pat_lru_tree && !LIST_ISEMPTY(&expr->patterns) && expr->ref->entry_cnt >= 20) {
 		unsigned long long seed = pat_lru_seed ^ (long)expr;
 
 		lru = lru64_get(XXH3(smp->data.u.str.area, smp->data.u.str.data, seed),
@@ -539,7 +539,7 @@ struct pattern *pat_match_bin(struct sample *smp, struct pattern_expr *expr, int
 	struct pattern *ret = NULL;
 	struct lru64 *lru = NULL;
 
-	if (pat_lru_tree && !LIST_ISEMPTY(&expr->patterns)) {
+	if (pat_lru_tree && !LIST_ISEMPTY(&expr->patterns) && expr->ref->entry_cnt >= 20) {
 		unsigned long long seed = pat_lru_seed ^ (long)expr;
 
 		lru = lru64_get(XXH3(smp->data.u.str.area, smp->data.u.str.data, seed),
@@ -608,7 +608,7 @@ struct pattern *pat_match_reg(struct sample *smp, struct pattern_expr *expr, int
 	struct pattern *ret = NULL;
 	struct lru64 *lru = NULL;
 
-	if (pat_lru_tree && !LIST_ISEMPTY(&expr->patterns)) {
+	if (pat_lru_tree && !LIST_ISEMPTY(&expr->patterns) && expr->ref->entry_cnt >= 5) {
 		unsigned long long seed = pat_lru_seed ^ (long)expr;
 
 		lru = lru64_get(XXH3(smp->data.u.str.area, smp->data.u.str.data, seed),
@@ -674,7 +674,7 @@ struct pattern *pat_match_beg(struct sample *smp, struct pattern_expr *expr, int
 	}
 
 	/* look in the list */
-	if (pat_lru_tree && !LIST_ISEMPTY(&expr->patterns)) {
+	if (pat_lru_tree && !LIST_ISEMPTY(&expr->patterns) && expr->ref->entry_cnt >= 20) {
 		unsigned long long seed = pat_lru_seed ^ (long)expr;
 
 		lru = lru64_get(XXH3(smp->data.u.str.area, smp->data.u.str.data, seed),
@@ -718,7 +718,7 @@ struct pattern *pat_match_end(struct sample *smp, struct pattern_expr *expr, int
 	struct pattern *ret = NULL;
 	struct lru64 *lru = NULL;
 
-	if (pat_lru_tree && !LIST_ISEMPTY(&expr->patterns)) {
+	if (pat_lru_tree && !LIST_ISEMPTY(&expr->patterns) && expr->ref->entry_cnt >= 20) {
 		unsigned long long seed = pat_lru_seed ^ (long)expr;
 
 		lru = lru64_get(XXH3(smp->data.u.str.area, smp->data.u.str.data, seed),
@@ -766,7 +766,7 @@ struct pattern *pat_match_sub(struct sample *smp, struct pattern_expr *expr, int
 	struct pattern *ret = NULL;
 	struct lru64 *lru = NULL;
 
-	if (pat_lru_tree && !LIST_ISEMPTY(&expr->patterns)) {
+	if (pat_lru_tree && !LIST_ISEMPTY(&expr->patterns) && expr->ref->entry_cnt >= 20) {
 		unsigned long long seed = pat_lru_seed ^ (long)expr;
 
 		lru = lru64_get(XXH3(smp->data.u.str.area, smp->data.u.str.data, seed),
@@ -1533,11 +1533,9 @@ struct pat_ref *pat_ref_lookup(const char *reference)
 {
 	struct pat_ref *ref;
 
-	/* Skip file@ or opt@ prefix, it is the default case. Can be mixed with ref omitting the prefix */
+	/* Skip file@ prefix, it is the default case. Can be mixed with ref omitting the prefix */
 	if (strlen(reference) > 5 && strncmp(reference, "file@", 5) == 0)
 		reference += 5;
-	else if (strlen(reference) > 4 && strncmp(reference, "opt@", 4) == 0)
-		reference += 4;
 
 	list_for_each_entry(ref, &pattern_reference, list)
 		if (ref->reference && strcmp(reference, ref->reference) == 0)
@@ -1794,6 +1792,52 @@ int pat_ref_set(struct pat_ref *ref, const char *key, const char *value, char **
 	return 1;
 }
 
+/* helper function to create and initialize a generic pat_ref struct
+ *
+ * Returns the new struct on success and NULL on failure (memory allocation
+ * error)
+ */
+static struct pat_ref *_pat_ref_new(const char *display, unsigned int flags)
+{
+	struct pat_ref *ref;
+
+	ref = malloc(sizeof(*ref));
+	if (!ref)
+		return NULL;
+
+	/* don't forget to explicitly initialize all pat_ref struct members */
+
+	if (display) {
+		ref->display = strdup(display);
+		if (!ref->display) {
+			free(ref);
+			return NULL;
+		}
+	}
+
+	ref->reference = NULL;
+	ref->flags = flags;
+	ref->curr_gen = 0;
+        ref->next_gen = 0;
+	ref->unique_id = -1;
+	ref->revision = 0;
+	ref->entry_cnt = 0;
+	LIST_INIT(&ref->head);
+	ref->ebmb_root = EB_ROOT;
+	LIST_INIT(&ref->pat);
+	HA_RWLOCK_INIT(&ref->lock);
+
+	return ref;
+}
+
+/* helper func to properly de-initialize and free pat_ref struct */
+static void pat_ref_free(struct pat_ref *ref)
+{
+	ha_free(&ref->reference);
+	ha_free(&ref->display);
+	free(ref);
+}
+
 /* This function creates a new reference. <ref> is the reference name.
  * <flags> are PAT_REF_*. /!\ The reference is not checked, and must
  * be unique. The user must check the reference with "pat_ref_lookup()"
@@ -1804,28 +1848,19 @@ struct pat_ref *pat_ref_new(const char *reference, const char *display, unsigned
 {
 	struct pat_ref *ref;
 
-	ref = calloc(1, sizeof(*ref));
+	ref = _pat_ref_new(display, flags);
 	if (!ref)
 		return NULL;
 
-	if (display) {
-		ref->display = strdup(display);
-		if (!ref->display) {
-			free(ref);
-			return NULL;
-		}
-	}
-
-
 	if (strlen(reference) > 5 && strncmp(reference, "virt@", 5) == 0)
-		flags |= PAT_REF_ID;
+		ref->flags |= PAT_REF_ID;
 	else if (strlen(reference) > 4 && strncmp(reference, "opt@", 4) == 0) {
-		flags |= (PAT_REF_ID|PAT_REF_FILE); // Will be decided later
+		ref->flags |= (PAT_REF_ID|PAT_REF_FILE); // Will be decided later
 		reference += 4;
 	}
 	else {
 		/* A file by default */
-		flags |= PAT_REF_FILE;
+		ref->flags |= PAT_REF_FILE;
 		/* Skip file@ prefix to be mixed with ref omitting the prefix */
 		if (strlen(reference) > 5 && strncmp(reference, "file@", 5) == 0)
 			reference += 5;
@@ -1834,22 +1869,11 @@ struct pat_ref *pat_ref_new(const char *reference, const char *display, unsigned
 
 	ref->reference = strdup(reference);
 	if (!ref->reference) {
-		free(ref->display);
-		free(ref);
+		pat_ref_free(ref);
 		return NULL;
 	}
 
-	ref->flags = flags;
-	ref->unique_id = -1;
-	ref->revision = 0;
-	ref->entry_cnt = 0;
-
-	LIST_INIT(&ref->head);
-	ref->ebmb_root = EB_ROOT;
-	LIST_INIT(&ref->pat);
-	HA_RWLOCK_INIT(&ref->lock);
 	LIST_APPEND(&pattern_reference, &ref->list);
-
 	return ref;
 }
 
@@ -1864,29 +1888,13 @@ struct pat_ref *pat_ref_newid(int unique_id, const char *display, unsigned int f
 {
 	struct pat_ref *ref;
 
-	ref = calloc(1, sizeof(*ref));
+	ref = _pat_ref_new(display, flags);
 	if (!ref)
 		return NULL;
 
-	if (display) {
-		ref->display = strdup(display);
-		if (!ref->display) {
-			free(ref);
-			return NULL;
-		}
-	}
-
-	ref->reference = NULL;
-	ref->flags = flags;
-	ref->curr_gen = 0;
-	ref->next_gen = 0;
 	ref->unique_id = unique_id;
-	LIST_INIT(&ref->head);
-	ref->ebmb_root = EB_ROOT;
-	LIST_INIT(&ref->pat);
-	HA_RWLOCK_INIT(&ref->lock);
-	LIST_APPEND(&pattern_reference, &ref->list);
 
+	LIST_APPEND(&pattern_reference, &ref->list);
 	return ref;
 }
 
@@ -2421,8 +2429,6 @@ int pattern_read_from_file(struct pattern_head *head, unsigned int refflags,
 					return 0;
 			}
 		}
-		else if ((ref->flags & PAT_REF_ID) && load_smp)
-			ref->flags |= PAT_REF_SMP;
 	}
 	else {
 		/* The reference already exists, check the map compatibility. */
