@@ -83,6 +83,8 @@ struct cubic {
 	uint32_t recovery_start_time;
 	/* HyStart++ state. */
 	struct quic_hystart hystart;
+	/* Consecutive number of losses since last ACK */
+	uint32_t consecutive_losses;
 };
 
 static void quic_cc_cubic_reset(struct quic_cc *cc)
@@ -107,7 +109,10 @@ static void quic_cc_cubic_reset(struct quic_cc *cc)
 
 static int quic_cc_cubic_init(struct quic_cc *cc)
 {
+	struct cubic *c = quic_cc_priv(cc);
+
 	quic_cc_cubic_reset(cc);
+	c->consecutive_losses = 0;
 	return 1;
 }
 
@@ -377,7 +382,11 @@ static inline void quic_cubic_update(struct quic_cc *cc, uint32_t acked)
 			inc = W_est_inc;
 	}
 
-	quic_cc_path_inc(path, inc);
+	if (quic_cwnd_may_increase(path)) {
+		path->cwnd += inc;
+		path->cwnd = QUIC_MIN(path->max_cwnd, path->cwnd);
+		path->mcwnd = QUIC_MAX(path->cwnd, path->mcwnd);
+	}
  leave:
 	TRACE_LEAVE(QUIC_EV_CONN_CC, cc->qc);
 }
@@ -424,7 +433,7 @@ static void quic_enter_recovery(struct quic_cc *cc)
 	}
 
 	c->ssthresh = (CUBIC_BETA_SCALED * path->cwnd) >> CUBIC_SCALE_FACTOR_SHIFT;
-	quic_cc_path_set(path, c->ssthresh);
+	path->cwnd =  QUIC_MAX(c->ssthresh, (uint32_t)path->min_cwnd);
 	c->state = QUIC_CC_ST_RP;
 	TRACE_LEAVE(QUIC_EV_CONN_CC, cc->qc, NULL, cc);
 }
@@ -446,7 +455,10 @@ static void quic_cc_cubic_ss_cb(struct quic_cc *cc, struct quic_cc_event *ev)
 			if (path->cwnd >= QUIC_CC_INFINITE_SSTHESH - acked)
 				goto out;
 
-			quic_cc_path_inc(path, acked);
+			if (quic_cwnd_may_increase(path)) {
+				path->cwnd += acked;
+				path->mcwnd = QUIC_MAX(path->cwnd, path->mcwnd);
+			}
 			quic_cc_hystart_track_min_rtt(cc, h, path->loss.latest_rtt);
 			if (ev->ack.pn >= h->wnd_end)
 				h->wnd_end = UINT64_MAX;
@@ -457,11 +469,15 @@ static void quic_cc_cubic_ss_cb(struct quic_cc *cc, struct quic_cc_event *ev)
 			}
 		}
 		else if (path->cwnd < QUIC_CC_INFINITE_SSTHESH - ev->ack.acked) {
-			quic_cc_path_inc(path, ev->ack.acked);
+			if (quic_cwnd_may_increase(path)) {
+				path->cwnd += ev->ack.acked;
+				path->cwnd = QUIC_MIN(path->max_cwnd, path->cwnd);
+			}
 		}
 		/* Exit to congestion avoidance if slow start threshold is reached. */
 		if (path->cwnd >= c->ssthresh)
 			c->state = QUIC_CC_ST_CA;
+		path->mcwnd = QUIC_MAX(path->cwnd, path->mcwnd);
 		break;
 
 	case QUIC_CC_EVT_LOSS:
@@ -481,13 +497,28 @@ static void quic_cc_cubic_ss_cb(struct quic_cc *cc, struct quic_cc_event *ev)
 /* Congestion avoidance callback. */
 static void quic_cc_cubic_ca_cb(struct quic_cc *cc, struct quic_cc_event *ev)
 {
+	struct cubic *c = quic_cc_priv(cc);
+
 	TRACE_ENTER(QUIC_EV_CONN_CC, cc->qc);
 	TRACE_PROTO("CC cubic", QUIC_EV_CONN_CC, cc->qc, ev);
 	switch (ev->type) {
 	case QUIC_CC_EVT_ACK:
+		c->consecutive_losses = 0;
 		quic_cubic_update(cc, ev->ack.acked);
 		break;
 	case QUIC_CC_EVT_LOSS:
+		/* Principle: we may want to tolerate one or a few occasional
+		 * losses that are *not* caused by congestion that we'd have
+		 * any control on. Tests show that over long distances this
+		 * significantly improves the transfer stability and
+		 * performance, but can quickly result in a massive loss
+		 * increase if set too high. This counter is reset upon ACKs.
+		 * Maybe we could refine this to consider only certain ACKs
+		 * though.
+		 */
+		c->consecutive_losses += ev->loss.count;
+		if (c->consecutive_losses <= global.tune.quic_cubic_loss_tol)
+			goto out;
 		quic_enter_recovery(cc);
 		break;
 	case QUIC_CC_EVT_ECN_CE:
@@ -519,7 +550,10 @@ static void quic_cc_cubic_cs_cb(struct quic_cc *cc, struct quic_cc_event *ev)
 		if (path->cwnd >= QUIC_CC_INFINITE_SSTHESH - acked)
 			goto out;
 
-		quic_cc_path_inc(path, acked);
+		if (quic_cwnd_may_increase(path)) {
+			path->cwnd += acked;
+			path->mcwnd = QUIC_MAX(path->cwnd, path->mcwnd);
+		}
 		quic_cc_hystart_track_min_rtt(cc, h, path->loss.latest_rtt);
 		if (quic_cc_hystart_may_reenter_ss(h)) {
 			/* Exit to slow start */
@@ -628,10 +662,10 @@ static void quic_cc_cubic_state_trace(struct buffer *buf, const struct quic_cc *
 	struct cubic *c = quic_cc_priv(cc);
 
 	path = container_of(cc, struct quic_cc_path, cc);
-	chunk_appendf(buf, " state=%s cwnd=%llu cwnd_last_max=%llu ssthresh=%d rpst=%dms",
+	chunk_appendf(buf, " state=%s cwnd=%llu mcwnd=%llu ssthresh=%d rpst=%dms",
 	              quic_cc_state_str(c->state),
 	              (unsigned long long)path->cwnd,
-	              (unsigned long long)path->cwnd_last_max,
+	              (unsigned long long)path->mcwnd,
 	              (int)c->ssthresh,
 	              !tick_isset(c->recovery_start_time) ? -1 :
 	              TICKS_TO_MS(tick_remain(c->recovery_start_time, now_ms)));
@@ -648,6 +682,7 @@ static void quic_cc_cubic_state_cli(struct buffer *buf, const struct quic_cc_pat
 
 struct quic_cc_algo quic_cc_algo_cubic = {
 	.type        = QUIC_CC_ALGO_TP_CUBIC,
+	.flags       = QUIC_CC_ALGO_FL_OPT_PACING,
 	.init        = quic_cc_cubic_init,
 	.event       = quic_cc_cubic_event,
 	.slow_start  = quic_cc_cubic_slow_start,
@@ -655,3 +690,11 @@ struct quic_cc_algo quic_cc_algo_cubic = {
 	.state_trace = quic_cc_cubic_state_trace,
 	.state_cli   = quic_cc_cubic_state_cli,
 };
+
+void quic_cc_cubic_check(void)
+{
+	struct quic_cc *cc;
+	BUG_ON_HOT(sizeof(struct cubic) > sizeof(cc->priv));
+}
+
+INITCALL0(STG_REGISTER, quic_cc_cubic_check);
