@@ -16,12 +16,14 @@
 
 #include <haproxy/h3.h>
 #include <haproxy/list.h>
-#include <haproxy/ncbmbuf.h>
+#include <haproxy/ncbuf.h>
 #include <haproxy/proto_quic.h>
 #include <haproxy/quic_ack.h>
+#include <haproxy/quic_cc_drs.h>
 #include <haproxy/quic_cid.h>
 #include <haproxy/quic_retransmit.h>
 #include <haproxy/quic_retry.h>
+#include <haproxy/quic_rules.h>
 #include <haproxy/quic_sock.h>
 #include <haproxy/quic_stream.h>
 #include <haproxy/quic_ssl.h>
@@ -92,7 +94,7 @@ static int qc_do_rm_hp(struct quic_conn *qc,
 
 	sample = pn + QUIC_PACKET_PN_MAXLEN;
 
-	if (!quic_tls_aes_decrypt(mask, sample, sizeof mask, tls_ctx->rx.hp_ctx)) {
+	if (!quic_tls_hp_decrypt(mask, sample, sizeof mask, tls_ctx->rx.hp_ctx, tls_ctx->rx.hp_key)) {
 		TRACE_ERROR("HP removing failed", QUIC_EV_CONN_RMHP, qc, pkt);
 		goto leave;
 	}
@@ -200,69 +202,15 @@ static int qc_pkt_decrypt(struct quic_conn *qc, struct quic_enc_level *qel,
 	return ret;
 }
 
-/* Remove from <stream> the acknowledged frames.
- *
- * Returns 1 if at least one frame was removed else 0.
- */
-static int quic_stream_try_to_consume(struct quic_conn *qc,
-                                      struct qc_stream_desc *stream)
-{
-	int ret;
-	struct eb64_node *frm_node;
-
-	TRACE_ENTER(QUIC_EV_CONN_ACKSTRM, qc);
-
-	ret = 0;
-	frm_node = eb64_first(&stream->acked_frms);
-	while (frm_node) {
-		struct qf_stream *strm_frm;
-		struct quic_frame *frm;
-		size_t offset, len;
-		int fin;
-
-		strm_frm = eb64_entry(frm_node, struct qf_stream, offset);
-		frm = container_of(strm_frm, struct quic_frame, stream);
-		offset = strm_frm->offset.key;
-		len = strm_frm->len;
-		fin = frm->type & QUIC_STREAM_FRAME_TYPE_FIN_BIT;
-
-		if (offset > stream->ack_offset)
-			break;
-
-		if (qc_stream_desc_ack(&stream, offset, len, fin)) {
-			/* cf. next comment : frame may be freed at this stage. */
-			TRACE_DEVEL("stream consumed", QUIC_EV_CONN_ACKSTRM,
-			            qc, stream ? strm_frm : NULL, stream);
-			ret = 1;
-		}
-
-		/* If stream is NULL after qc_stream_desc_ack(), it means frame
-		 * has been freed. with the stream frames tree. Nothing to do
-		 * anymore in here.
-		 */
-		if (!stream) {
-			qc_check_close_on_released_mux(qc);
-			ret = 1;
-			goto leave;
-		}
-
-		frm_node = eb64_next(frm_node);
-		eb64_delete(&strm_frm->offset);
-
-		qc_release_frm(qc, frm);
-	}
-
- leave:
-	TRACE_LEAVE(QUIC_EV_CONN_ACKSTRM, qc);
-	return ret;
-}
-
 /* Handle <frm> frame whose packet it is attached to has just been acknowledged. The memory allocated
  * for this frame will be at least released in every cases.
- * Never fail.
+ *
+ * Returns 1 on sucess else 0.
  */
-static void qc_handle_newly_acked_frm(struct quic_conn *qc, struct quic_frame *frm)
+static int qc_handle_newly_acked_frm(struct quic_conn *qc, struct quic_frame *frm)
 {
+	int ret = 0;
+
 	TRACE_ENTER(QUIC_EV_CONN_PRSAFRM, qc);
 	TRACE_PROTO("RX ack TX frm", QUIC_EV_CONN_PRSAFRM, qc, frm);
 
@@ -272,9 +220,7 @@ static void qc_handle_newly_acked_frm(struct quic_conn *qc, struct quic_frame *f
 		struct qf_stream *strm_frm = &frm->stream;
 		struct eb64_node *node = NULL;
 		struct qc_stream_desc *stream = NULL;
-		const size_t offset = strm_frm->offset.key;
-		const size_t len = strm_frm->len;
-		const int fin = frm->type & QUIC_STREAM_FRAME_TYPE_FIN_BIT;
+		int ack;
 
 		/* do not use strm_frm->stream as the qc_stream_desc instance
 		 * might be freed at this stage. Use the id to do a proper
@@ -287,43 +233,46 @@ static void qc_handle_newly_acked_frm(struct quic_conn *qc, struct quic_frame *f
 		if (!node) {
 			TRACE_DEVEL("acked stream for released stream", QUIC_EV_CONN_ACKSTRM, qc, strm_frm);
 			qc_release_frm(qc, frm);
-			/* early return */
-			goto leave;
-		}
-		stream = eb64_entry(node, struct qc_stream_desc, by_id);
-
-		TRACE_DEVEL("acked stream", QUIC_EV_CONN_ACKSTRM, qc, strm_frm, stream);
-		if (offset <= stream->ack_offset) {
-			if (qc_stream_desc_ack(&stream, offset, len, fin)) {
-				TRACE_DEVEL("stream consumed", QUIC_EV_CONN_ACKSTRM,
-				            qc, strm_frm, stream);
-			}
-
-			if (!stream) {
-				/* no need to continue if stream freed. */
-				TRACE_DEVEL("stream released and freed", QUIC_EV_CONN_ACKSTRM, qc);
-				qc_release_frm(qc, frm);
-				qc_check_close_on_released_mux(qc);
-				break;
-			}
-
-			TRACE_DEVEL("stream consumed", QUIC_EV_CONN_ACKSTRM,
-			            qc, strm_frm, stream);
-			qc_release_frm(qc, frm);
+			/* return as success */
 		}
 		else {
-			eb64_insert(&stream->acked_frms, &strm_frm->offset);
-		}
+			stream = eb64_entry(node, struct qc_stream_desc, by_id);
 
-		quic_stream_try_to_consume(qc, stream);
+			ack = qc_stream_desc_ack(stream, strm_frm->offset,
+			                         strm_frm->len,
+			                         frm->type & QUIC_STREAM_FRAME_TYPE_FIN_BIT);
+			if (!ack) {
+				TRACE_DEVEL("stream consumed on ACK received",
+				            QUIC_EV_CONN_ACKSTRM, qc, strm_frm, stream);
+
+				if (qc_stream_desc_done(stream)) {
+					/* no need to continue if stream freed. */
+					TRACE_DEVEL("stream released and freed", QUIC_EV_CONN_ACKSTRM, qc);
+					qc_check_close_on_released_mux(qc);
+				}
+
+				qc_release_frm(qc, frm);
+			}
+			else if (ack > 0) {
+				TRACE_DEVEL("handled out-of-order stream ACK",
+				            QUIC_EV_CONN_ACKSTRM, qc, strm_frm, stream);
+				qc_release_frm(qc, frm);
+			}
+			else {
+				/* Fatal error during qc_stream_desc_ack(). */
+				goto leave;
+			}
+		}
 	}
 	break;
 	default:
 		qc_release_frm(qc, frm);
 	}
 
+	ret = 1;
  leave:
 	TRACE_LEAVE(QUIC_EV_CONN_PRSAFRM, qc);
+	return ret;
 }
 
 /* Collect newly acknowledged TX packets from <pkts> ebtree into <newly_acked_pkts>
@@ -360,11 +309,15 @@ static void qc_newly_acked_pkts(struct quic_conn *qc, struct eb_root *pkts,
 	TRACE_LEAVE(QUIC_EV_CONN_PRSAFRM, qc);
 }
 
-/* Handle <newly_acked_pkts> list of newly acknowledged TX packets */
-static void qc_handle_newly_acked_pkts(struct quic_conn *qc,
-                                       unsigned int *pkt_flags, struct list *newly_acked_pkts)
+/* Handle <newly_acked_pkts> list of newly acknowledged TX packets.
+ *
+ * Returns 1 on sucess else 0.
+ */
+static int qc_handle_newly_acked_pkts(struct quic_conn *qc,
+                                      unsigned int *pkt_flags, struct list *newly_acked_pkts)
 {
 	struct quic_tx_packet *pkt, *tmp;
+	int ret = 0;
 
 	TRACE_ENTER(QUIC_EV_CONN_PRSAFRM, qc);
 
@@ -373,8 +326,10 @@ static void qc_handle_newly_acked_pkts(struct quic_conn *qc,
 
 		*pkt_flags |= pkt->flags;
 		TRACE_DEVEL("Removing packet #", QUIC_EV_CONN_PRSAFRM, qc, NULL, &pkt->pn_node.key);
-		list_for_each_entry_safe(frm, frmbak, &pkt->frms, list)
-			qc_handle_newly_acked_frm(qc, frm);
+		list_for_each_entry_safe(frm, frmbak, &pkt->frms, list) {
+			if (!qc_handle_newly_acked_frm(qc, frm))
+				goto leave;
+		}
 		/* If there are others packet in the same datagram <pkt> is attached to,
 		 * detach the previous one and the next one from <pkt>.
 		 */
@@ -382,8 +337,10 @@ static void qc_handle_newly_acked_pkts(struct quic_conn *qc,
 		eb64_delete(&pkt->pn_node);
 	}
 
+	ret = 1;
  leave:
 	TRACE_LEAVE(QUIC_EV_CONN_PRSAFRM, qc);
+	return ret;
 }
 
 /* Handle all frames sent from <pkt> packet and reinsert them in the same order
@@ -465,33 +422,53 @@ int qc_handle_frms_of_lost_pkt(struct quic_conn *qc,
  * Always succeeds.
  */
 static void qc_notify_cc_of_newly_acked_pkts(struct quic_conn *qc,
-                                             struct list *newly_acked_pkts)
+                                             struct list *newly_acked_pkts,
+                                             unsigned int bytes_lost,
+                                             unsigned int rtt)
 {
 	struct quic_tx_packet *pkt, *tmp;
 	struct quic_cc_event ev = { .type = QUIC_CC_EVT_ACK, };
+	struct quic_cc_path *p = qc->path;
+	struct quic_cc_drs *drs =
+		p->cc.algo->get_drs ? p->cc.algo->get_drs(&p->cc) : NULL;
+	unsigned int bytes_delivered = 0, pkt_delivered = 0;
+	uint64_t time_ns = task_mono_time();
 
 	TRACE_ENTER(QUIC_EV_CONN_PRSAFRM, qc);
 
 	list_for_each_entry_safe(pkt, tmp, newly_acked_pkts, list) {
 		pkt->pktns->tx.in_flight -= pkt->in_flight_len;
-		qc->path->prep_in_flight -= pkt->in_flight_len;
+		p->prep_in_flight -= pkt->in_flight_len;
 		if (pkt->flags & QUIC_FL_TX_PACKET_ACK_ELICITING)
-			qc->path->ifae_pkts--;
+			p->ifae_pkts--;
 		/* If this packet contained an ACK frame, proceed to the
 		 * acknowledging of range of acks from the largest acknowledged
 		 * packet number which was sent in an ACK frame by this packet.
 		 */
 		if (pkt->largest_acked_pn != -1)
 			qc_treat_ack_of_ack(qc, &pkt->pktns->rx.arngs, pkt->largest_acked_pn);
+		bytes_delivered += pkt->len;
+		pkt_delivered = pkt->rs.delivered;
 		ev.ack.acked = pkt->in_flight_len;
-		ev.ack.time_sent = pkt->time_sent;
+		ev.ack.time_sent = pkt->time_sent_ms;
 		ev.ack.pn = pkt->pn_node.key;
-		quic_cc_event(&qc->path->cc, &ev);
-		qc->path->in_flight -= pkt->in_flight_len;
+		/* Note that this event is not emitted for BBR. */
+		quic_cc_event(&p->cc, &ev);
+		p->in_flight -= pkt->in_flight_len;
+		if (drs && (pkt->flags & QUIC_FL_TX_PACKET_ACK_ELICITING))
+			quic_cc_drs_update_rate_sample(drs, pkt, time_ns);
 		LIST_DEL_INIT(&pkt->list);
 		quic_tx_packet_refdec(pkt);
 	}
 
+	if (drs) {
+		quic_cc_drs_on_ack_recv(drs, p, pkt_delivered);
+		drs->lost += bytes_lost;
+	}
+	if (p->cc.algo->on_ack_rcvd)
+		p->cc.algo->on_ack_rcvd(&p->cc, bytes_delivered, pkt_delivered,
+		                        rtt, bytes_lost, now_ms);
+
 	TRACE_LEAVE(QUIC_EV_CONN_PRSAFRM, qc);
 
 }
@@ -547,7 +524,7 @@ static int qc_parse_ack_frm(struct quic_conn *qc,
 		}
 		else {
 			time_sent = eb64_entry(largest_node,
-			                       struct quic_tx_packet, pn_node)->time_sent;
+			                       struct quic_tx_packet, pn_node)->time_sent_ms;
 			new_largest_acked_pn = 1;
 		}
 	}
@@ -595,18 +572,24 @@ static int qc_parse_ack_frm(struct quic_conn *qc,
 	} while (1);
 
 	if (!LIST_ISEMPTY(&newly_acked_pkts)) {
-		qc_handle_newly_acked_pkts(qc, &pkt_flags, &newly_acked_pkts);
+		unsigned int bytes_lost = 0;
+
+		if (!qc_handle_newly_acked_pkts(qc, &pkt_flags, &newly_acked_pkts))
+			goto leave;
+
 		if (new_largest_acked_pn && (pkt_flags & QUIC_FL_TX_PACKET_ACK_ELICITING)) {
 			*rtt_sample = tick_remain(time_sent, now_ms);
 			qel->pktns->rx.largest_acked_pn = ack_frm->largest_ack;
 		}
 
 		if (!eb_is_empty(&qel->pktns->tx.pkts)) {
-			qc_packet_loss_lookup(qel->pktns, qc, &lost_pkts);
+			qc_packet_loss_lookup(qel->pktns, qc, &lost_pkts, &bytes_lost);
 			if (!qc_release_lost_pkts(qc, qel->pktns, &lost_pkts, now_ms))
 				goto leave;
 		}
-		qc_notify_cc_of_newly_acked_pkts(qc, &newly_acked_pkts);
+
+		qc_notify_cc_of_newly_acked_pkts(qc, &newly_acked_pkts,
+		                                 bytes_lost, *rtt_sample);
 		if (quic_peer_validated_addr(qc))
 			qc->path->loss.pto_count = 0;
 		qc_set_timer(qc);
@@ -650,7 +633,7 @@ static int qc_handle_strm_frm(struct quic_rx_packet *pkt,
 	TRACE_ENTER(QUIC_EV_CONN_PRSFRM, qc);
 
 	ret = qcc_recv(qc->qcc, strm_frm->id, strm_frm->len,
-	               strm_frm->offset.key, fin, (char *)strm_frm->data);
+	               strm_frm->offset, fin, (char *)strm_frm->data);
 
 	/* frame rejected - packet must not be acknowledeged */
 	TRACE_LEAVE(QUIC_EV_CONN_PRSFRM, qc);
@@ -675,19 +658,19 @@ static enum quic_rx_ret_frm qc_handle_crypto_frm(struct quic_conn *qc,
 	enum quic_rx_ret_frm ret = QUIC_RX_RET_FRM_DONE;
 	/* XXX TO DO: <cfdebug> is used only for the traces. */
 	struct quic_rx_crypto_frm cfdebug = {
-		.offset_node.key = crypto_frm->offset_node.key,
+		.offset_node.key = crypto_frm->offset,
 		.len = crypto_frm->len,
 	};
 	struct quic_cstream *cstream = qel->cstream;
-	struct ncbmbuf *ncbuf = &qel->cstream->rx.ncbuf;
+	struct ncbuf *ncbuf = &qel->cstream->rx.ncbuf;
 	uint64_t off_rel;
 
 	TRACE_ENTER(QUIC_EV_CONN_PRSHPKT, qc);
 
-	if (unlikely(crypto_frm->offset_node.key < cstream->rx.offset)) {
+	if (unlikely(crypto_frm->offset < cstream->rx.offset)) {
 		size_t diff;
 
-		if (crypto_frm->offset_node.key + crypto_frm->len <= cstream->rx.offset) {
+		if (crypto_frm->offset + crypto_frm->len <= cstream->rx.offset) {
 			/* Nothing to do */
 			TRACE_PROTO("Already received CRYPTO data",
 			            QUIC_EV_CONN_RXPKT, qc, pkt, &cfdebug);
@@ -698,19 +681,19 @@ static enum quic_rx_ret_frm qc_handle_crypto_frm(struct quic_conn *qc,
 		TRACE_PROTO("Partially already received CRYPTO data",
 		            QUIC_EV_CONN_RXPKT, qc, pkt, &cfdebug);
 
-		diff = cstream->rx.offset - crypto_frm->offset_node.key;
+		diff = cstream->rx.offset - crypto_frm->offset;
 		crypto_frm->len -= diff;
 		crypto_frm->data += diff;
-		crypto_frm->offset_node.key = cstream->rx.offset;
+		crypto_frm->offset = cstream->rx.offset;
 	}
 
-	if (!quic_get_ncbuf(ncbuf) || ncbmb_is_null(ncbuf)) {
+	if (!quic_get_ncbuf(ncbuf) || ncb_is_null(ncbuf)) {
 		TRACE_ERROR("CRYPTO ncbuf allocation failed", QUIC_EV_CONN_PRSHPKT, qc);
 		goto err;
 	}
 
 	/* crypto_frm->offset > cstream-trx.offset */
-	off_rel = crypto_frm->offset_node.key - cstream->rx.offset;
+	off_rel = crypto_frm->offset - cstream->rx.offset;
 
 	/* RFC 9000 7.5. Cryptographic Message Buffering
 	 *
@@ -720,18 +703,31 @@ static enum quic_rx_ret_frm qc_handle_crypto_frm(struct quic_conn *qc,
 	 * handshake. If an endpoint does not expand its buffer, it MUST close
 	 * the connection with a CRYPTO_BUFFER_EXCEEDED error code.
 	 */
-	if (off_rel + crypto_frm->len > ncbmb_size(ncbuf)) {
+	if (off_rel + crypto_frm->len > ncb_size(ncbuf)) {
 		TRACE_ERROR("CRYPTO frame too large", QUIC_EV_CONN_PRSHPKT, qc);
 		quic_set_connection_close(qc, quic_err_transport(QC_ERR_CRYPTO_BUFFER_EXCEEDED));
 		goto err;
 	}
 
-	ncb_ret = ncbmb_add(ncbuf, off_rel, (const char *)crypto_frm->data,
-	                    crypto_frm->len, NCB_ADD_OVERWRT);
-	BUG_ON(ncb_ret != NCB_RET_OK);
+	ncb_ret = ncb_add(ncbuf, off_rel, (const char *)crypto_frm->data,
+	                  crypto_frm->len, NCB_ADD_COMPARE);
+	if (ncb_ret != NCB_RET_OK) {
+		if (ncb_ret == NCB_RET_DATA_REJ) {
+			TRACE_ERROR("overlapping data rejected", QUIC_EV_CONN_PRSHPKT, qc);
+			quic_set_connection_close(qc, quic_err_transport(QC_ERR_PROTOCOL_VIOLATION));
+			qc_notify_err(qc);
+			goto err;
+		}
+		else if (ncb_ret == NCB_RET_GAP_SIZE) {
+			TRACE_DATA("cannot bufferize frame due to gap size limit",
+			           QUIC_EV_CONN_PRSHPKT, qc);
+			ret = QUIC_RX_RET_FRM_AGAIN;
+			goto done;
+		}
+	}
 
 	/* Reschedule with TASK_HEAVY if CRYPTO data ready for decoding. */
-	if (ncbmb_data(ncbuf, 0)) {
+	if (ncb_data(ncbuf, 0)) {
 		HA_ATOMIC_OR(&qc->wait_event.tasklet->state, TASK_HEAVY);
 		tasklet_wakeup(qc->wait_event.tasklet);
 	}
@@ -820,12 +816,13 @@ static inline unsigned int quic_ack_delay_ms(struct qf_ack *ack_frm,
 static int qc_parse_pkt_frms(struct quic_conn *qc, struct quic_rx_packet *pkt,
                              struct quic_enc_level *qel)
 {
-	struct eb_root cf_frms_tree = EB_ROOT;
-	struct eb64_node *node;
-	struct quic_frame *frm = NULL;
+	struct list retry_frms = LIST_HEAD_INIT(retry_frms);
+	struct quic_frame *frm = NULL, *frm_tmp;
 	const unsigned char *pos, *end;
 	enum quic_rx_ret_frm ret;
-	int fast_retrans = 0, cf_err = 0;
+	int fast_retrans = 0;
+	/* parsing may be rerun multiple times, but no more than <iter>. */
+	int iter = 3, parsing_stage = 0;
 
 	TRACE_ENTER(QUIC_EV_CONN_PRSHPKT, qc);
 	/* Skip the AAD */
@@ -903,8 +900,36 @@ static int qc_parse_pkt_frms(struct quic_conn *qc, struct quic_rx_packet *pkt,
 			break;
 		}
 		case QUIC_FT_CRYPTO:
-			eb64_insert(&cf_frms_tree, &frm->crypto.offset_node);
-			frm = NULL;
+			ret = qc_handle_crypto_frm(qc, &frm->crypto, pkt, qel);
+			switch (ret) {
+			case QUIC_RX_RET_FRM_FATAL:
+				goto err;
+
+			case QUIC_RX_RET_FRM_AGAIN:
+				if (parsing_stage == 0) {
+					TRACE_STATE("parsing stage set to 1 (AGAIN encountered)", QUIC_EV_CONN_PRSHPKT, qc);
+					++parsing_stage;
+				}
+				/* Save frame in temp list to reparse it later. A new instance must be used for next packet frames. */
+				LIST_APPEND(&retry_frms, &frm->list);
+				frm = NULL;
+				break;
+
+			case QUIC_RX_RET_FRM_DUP:
+				if (qc_is_listener(qc) && qel == qc->iel &&
+				    !(qc->flags & QUIC_FL_CONN_HANDSHAKE_SPEED_UP)) {
+					fast_retrans = 1;
+				}
+				break;
+
+			case QUIC_RX_RET_FRM_DONE:
+				if (parsing_stage == 1) {
+					TRACE_STATE("parsing stage set to 2 (DONE after AGAIN)", QUIC_EV_CONN_PRSHPKT, qc);
+					++parsing_stage;
+				}
+				break;
+			}
+
 			break;
 		case QUIC_FT_NEW_TOKEN:
 			if (qc_is_listener(qc)) {
@@ -995,7 +1020,6 @@ static int qc_parse_pkt_frms(struct quic_conn *qc, struct quic_rx_packet *pkt,
 			break;
 		case QUIC_FT_RETIRE_CONNECTION_ID:
 		{
-			struct quic_cid_tree *tree __maybe_unused;
 			struct quic_connection_id *conn_id = NULL;
 
 			if (!qc_handle_retire_connection_id_frm(qc, frm, &pkt->dcid, &conn_id))
@@ -1004,10 +1028,7 @@ static int qc_parse_pkt_frms(struct quic_conn *qc, struct quic_rx_packet *pkt,
 			if (!conn_id)
 				break;
 
-			tree = &quic_cid_trees[quic_cid_tree_idx(&conn_id->cid)];
-			HA_RWLOCK_WRLOCK(QC_CID_LOCK, &tree->lock);
-			ebmb_delete(&conn_id->node);
-			HA_RWLOCK_WRUNLOCK(QC_CID_LOCK, &tree->lock);
+			quic_cid_delete(conn_id);
 			eb64_delete(&conn_id->seq_num);
 			pool_free(pool_head_quic_connection_id, conn_id);
 			TRACE_PROTO("CID retired", QUIC_EV_CONN_PSTRM, qc);
@@ -1015,12 +1036,9 @@ static int qc_parse_pkt_frms(struct quic_conn *qc, struct quic_rx_packet *pkt,
 			conn_id = new_quic_cid(qc->cids, qc, NULL, NULL);
 			if (!conn_id) {
 				TRACE_ERROR("CID allocation error", QUIC_EV_CONN_IO_CB, qc);
-				quic_set_connection_close(qc, quic_err_transport(QC_ERR_INTERNAL_ERROR));
-				qc_notify_err(qc);
-				goto err;
 			}
 			else {
-				quic_cid_insert(conn_id);
+				_quic_cid_insert(conn_id);
 				qc_build_new_connection_id_frm(qc, conn_id);
 			}
 			break;
@@ -1076,45 +1094,57 @@ static int qc_parse_pkt_frms(struct quic_conn *qc, struct quic_rx_packet *pkt,
 	if (frm)
 		qc_frm_free(qc, &frm);
 
-	node = eb64_first(&cf_frms_tree);
-	while (node) {
-		frm = eb64_entry(node, struct quic_frame, crypto.offset_node);
-		/* only CRYPTO frames may be reparsed for now */
-		BUG_ON(frm->type != QUIC_FT_CRYPTO);
-		node = eb64_next(node);
-		ret = qc_handle_crypto_frm(qc, &frm->crypto, pkt, qel);
-		switch (ret) {
-		case QUIC_RX_RET_FRM_FATAL:
-			/* avoid freeing without eb_delete() */
-			frm = NULL;
+	while (!LIST_ISEMPTY(&retry_frms)) {
+		if (--iter <= 0) {
+			TRACE_ERROR("interrupt parsing due to max iteration reached",
+			            QUIC_EV_CONN_PRSHPKT, qc);
+			goto err;
+		}
+		else if (parsing_stage <= 1) {
+			TRACE_ERROR("interrupt parsing due to buffering blocked on gap size limit",
+			            QUIC_EV_CONN_PRSHPKT, qc);
 			goto err;
-
-		case QUIC_RX_RET_FRM_AGAIN:
-			TRACE_STATE("AGAIN encountered", QUIC_EV_CONN_PRSHPKT, qc);
-			cf_err = 1;
-			break;
-
-		case QUIC_RX_RET_FRM_DONE:
-			TRACE_PROTO("frame handled", QUIC_EV_CONN_PRSAFRM, qc, frm);
-			break;
-
-		case QUIC_RX_RET_FRM_DUP:
-			if (qc_is_listener(qc) && qel == qc->iel &&
-				!(qc->flags & QUIC_FL_CONN_HANDSHAKE_SPEED_UP)) {
-				fast_retrans = 1;
-			}
-			break;
 		}
 
-		eb64_delete(&frm->crypto.offset_node);
-		qc_frm_free(qc, &frm);
+		parsing_stage = 0;
+		list_for_each_entry_safe(frm, frm_tmp, &retry_frms, list) {
+			/* only CRYPTO frames may be reparsed for now */
+			BUG_ON(frm->type != QUIC_FT_CRYPTO);
+			ret = qc_handle_crypto_frm(qc, &frm->crypto, pkt, qel);
+			switch (ret) {
+			case QUIC_RX_RET_FRM_FATAL:
+				goto err;
+
+			case QUIC_RX_RET_FRM_AGAIN:
+				if (parsing_stage == 0) {
+					TRACE_STATE("parsing stage set to 1 (AGAIN encountered)", QUIC_EV_CONN_PRSHPKT, qc);
+					++parsing_stage;
+				}
+				break;
+
+			case QUIC_RX_RET_FRM_DONE:
+				TRACE_PROTO("frame handled after a new parsing iteration",
+				            QUIC_EV_CONN_PRSAFRM, qc, frm);
+				if (parsing_stage == 1) {
+					TRACE_STATE("parsing stage set to 2 (DONE after AGAIN)", QUIC_EV_CONN_PRSHPKT, qc);
+					++parsing_stage;
+				}
+				__fallthrough;
+			case QUIC_RX_RET_FRM_DUP:
+				qc_frm_free(qc, &frm);
+				break;
+			}
+		}
+
+		/* Always reset <frm> as it may be dangling after
+		 * list_for_each_entry_safe() usage. Especially necessary to
+		 * prevent a crash if loop is interrupted on max iteration.
+		 */
+		frm = NULL;
 	}
 
-	if (cf_err)
-		goto err;
-
 	/* Error should be returned if some frames cannot be parsed. */
-	BUG_ON(!eb_is_empty(&cf_frms_tree));
+	BUG_ON(!LIST_ISEMPTY(&retry_frms));
 
 	if (fast_retrans && qc->iel && qc->hel) {
 		struct quic_enc_level *iqel = qc->iel;
@@ -1150,11 +1180,7 @@ static int qc_parse_pkt_frms(struct quic_conn *qc, struct quic_rx_packet *pkt,
  err:
 	if (frm)
 		qc_frm_free(qc, &frm);
-	node = eb64_first(&cf_frms_tree);
-	while (node) {
-		frm = eb64_entry(node, struct quic_frame, crypto.offset_node);
-		node = eb64_next(node);
-		eb64_delete(&frm->crypto.offset_node);
+	list_for_each_entry_safe(frm, frm_tmp, &retry_frms, list) {
 		qc_frm_free(qc, &frm);
 	}
 
@@ -1517,7 +1543,7 @@ static int qc_try_rm_hp(struct quic_conn *qc, struct quic_rx_packet *pkt,
 
 	ret = 1;
  out:
-	TRACE_LEAVE(QUIC_EV_CONN_TRMHP, qc, pkt);
+	TRACE_LEAVE(QUIC_EV_CONN_TRMHP, qc);
 	return ret;
 }
 
@@ -1687,6 +1713,9 @@ static struct quic_conn *quic_rx_pkt_retrieve_conn(struct quic_rx_packet *pkt,
 
 	qc = retrieve_qc_conn_from_cid(pkt, &dgram->saddr, new_tid);
 
+	/* quic_conn must be set to NULL if bind on another thread. */
+	BUG_ON_HOT(qc && *new_tid != -1);
+
 	/* If connection already created or rebinded on another thread. */
 	if (!qc && *new_tid != -1 && tid != *new_tid)
 		goto out;
@@ -1695,8 +1724,6 @@ static struct quic_conn *quic_rx_pkt_retrieve_conn(struct quic_rx_packet *pkt,
 		BUG_ON(!pkt->version); /* This must not happen. */
 
 		if (!qc) {
-			struct quic_cid_tree *tree;
-			struct ebmb_node *node;
 			struct quic_connection_id *conn_id;
 			int ipv4;
 
@@ -1735,13 +1762,24 @@ static struct quic_conn *quic_rx_pkt_retrieve_conn(struct quic_rx_packet *pkt,
 			}
 			else {
 				TRACE_PROTO("Initial without token", QUIC_EV_CONN_LPKT, NULL, NULL, NULL, pkt->version);
-				if (!(l->bind_conf->options & BC_O_QUIC_FORCE_RETRY) &&
-				    HA_ATOMIC_LOAD(&prx_counters->half_open_conn) >= global.tune.quic_retry_threshold) {
+			}
+
+			if (!quic_init_exec_rules(l, dgram)) {
+				TRACE_USER("drop datagram on quic-initial rules", QUIC_EV_CONN_LPKT, NULL, NULL, NULL, pkt->version);
+				goto err;
+			}
+
+			/* No need to emit Retry if connection is refused. */
+			if (!pkt->token_len && !(dgram->flags & QUIC_DGRAM_FL_REJECT)) {
+				if ((l->bind_conf->options & BC_O_QUIC_FORCE_RETRY) ||
+				    HA_ATOMIC_LOAD(&prx_counters->half_open_conn) >= global.tune.quic_retry_threshold ||
+				    (dgram->flags & QUIC_DGRAM_FL_SEND_RETRY)) {
+
 					TRACE_PROTO("Initial without token, sending retry",
-					            QUIC_EV_CONN_LPKT, NULL, NULL, NULL, pkt->version);
+						    QUIC_EV_CONN_LPKT, NULL, NULL, NULL, pkt->version);
 					if (send_retry(l->rx.fd, &dgram->saddr, pkt, pkt->version)) {
 						TRACE_ERROR("Error during Retry generation",
-									QUIC_EV_CONN_LPKT, NULL, NULL, NULL, pkt->version);
+							    QUIC_EV_CONN_LPKT, NULL, NULL, NULL, pkt->version);
 						goto out;
 					}
 
@@ -1788,14 +1826,8 @@ static struct quic_conn *quic_rx_pkt_retrieve_conn(struct quic_rx_packet *pkt,
 				qc->hash64 = quic_hash64_from_cid(conn_id->cid.data, conn_id->cid.len,
 								  global.cluster_secret, sizeof(global.cluster_secret));
 
-			tree = &quic_cid_trees[quic_cid_tree_idx(&conn_id->cid)];
-			HA_RWLOCK_WRLOCK(QC_CID_LOCK, &tree->lock);
-			node = ebmb_insert(&tree->root, &conn_id->node, conn_id->cid.len);
-			if (node != &conn_id->node) {
+			if (quic_cid_insert(conn_id, new_tid)) {
 				pool_free(pool_head_quic_connection_id, conn_id);
-
-				conn_id = ebmb_entry(node, struct quic_connection_id, node);
-				*new_tid = HA_ATOMIC_LOAD(&conn_id->tid);
 				quic_conn_release(qc);
 				qc = NULL;
 			}
@@ -1807,8 +1839,10 @@ static struct quic_conn *quic_rx_pkt_retrieve_conn(struct quic_rx_packet *pkt,
 				eb64_insert(qc->cids, &conn_id->seq_num);
 				/* Initialize the next CID sequence number to be used for this connection. */
 				qc->next_cid_seq_num = 1;
+
+				if (dgram->flags & QUIC_DGRAM_FL_REJECT)
+					quic_set_connection_close(qc, quic_err_transport(QC_ERR_CONNECTION_REFUSED));
 			}
-			HA_RWLOCK_WRUNLOCK(QC_CID_LOCK, &tree->lock);
 
 			if (*new_tid != -1)
 				goto out;
@@ -1937,24 +1971,6 @@ static int quic_rx_pkt_parse(struct quic_rx_packet *pkt,
 				goto drop;
 			}
 
-			/* TODO Retry should be automatically activated if
-			 * suspect network usage is detected.
-			 */
-			if (!token_len) {
-				if (l->bind_conf->options & BC_O_QUIC_FORCE_RETRY) {
-					TRACE_PROTO("Initial without token, sending retry",
-					            QUIC_EV_CONN_LPKT, NULL, NULL, NULL, pkt->version);
-					if (send_retry(l->rx.fd, &dgram->saddr, pkt, pkt->version)) {
-						TRACE_PROTO("Error during Retry generation",
-						            QUIC_EV_CONN_LPKT, NULL, NULL, NULL, pkt->version);
-						goto drop_silent;
-					}
-
-					HA_ATOMIC_INC(&prx_counters->retry_sent);
-					goto drop_silent;
-				}
-			}
-
 			pkt->token = pos;
 			pkt->token_len = token_len;
 			pos += pkt->token_len;
@@ -2302,6 +2318,10 @@ int quic_dgram_parse(struct quic_dgram *dgram, struct quic_conn *from_qc,
 			dgram->qc = qc;
 		}
 
+		/* Ensure quic_conn access only occurs on its attached thread. */
+		BUG_ON_HOT(((struct quic_connection_id *)
+		               eb64_entry(eb64_first(qc->cids), struct quic_connection_id, seq_num))->tid != tid);
+
 		/* Ensure thread connection migration is finalized ASAP. */
 		if (qc->flags & QUIC_FL_CONN_TID_REBIND)
 			qc_finalize_tid_rebind(qc);
